/*******************************************************************************\
 * Copyright (c) 1996, 1997 by Markus Koelblin
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
\*******************************************************************************/

/*******************************************************************************\
 * colors.c                                                                    *
 *                                                                             *
 * Implements X11 colormap routines - mainly for PseudoColor visuals           *
 *                                                                             *
 * Author:  Markus Koelblin (Markus.Koelblin@hadiko.de) 1996, 1997             *
\*******************************************************************************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "colors.h"

#define BITS_NONE (~0u)

/*************************************************************************\
 * Tool functions                                                        *
\*************************************************************************/

static unsigned first_bit_nr(unsigned n)
{
  /* Return the number of the first bit set */
  unsigned num;

  if (n == 0)
      return BITS_NONE;
  else
      for (num= 0; ((1 << num) & n) == 0; num++)
	  ;
  return num;
}

static int ln_2(unsigned n)
{
    int i;

    for (i = 0;  i < PLANES_MAX && n > 0;  i++)
	n = n >> 1;
    return i;
}

static int pow_2(int n)
{
    int i = 1;

    while (n-- > 0)
	i *= 2;
    return i;
}


/*************************************************************************\
 *  set_xcolor_rgb()                                                     *
 *                                                                       *
 *  Fills a XColor structure like the SET_XCOLOR_RGB() macro             *
\*************************************************************************/

void set_xcolor_rgb(XColor *color, unsigned short red, unsigned short green,
		    unsigned short blue)
{
    color->red   = red;
    color->green = green;
    color->blue  = blue;
    color->flags = DoRed | DoGreen | DoBlue;
}



/*************************************************************************\
 *  create_colormap()                                                    *
 *                                                                       *
 *  Creates a read/write colormap on PseudocColor displays, with         *
 *  n_pixels r/w color cells.                                            *
 *  Exisiting colors of the default colormap are copied if possible.     *
 *                                                                       *
 *  Returns:  A new colormap, and fills the pixels_return[]              *
 *            array with the allocated pixel values.                     *
\*************************************************************************/

#ifndef NDEBUG
#define PRINT_MAP_DEBUG \
	fputs("\n * create_colormap debug info *\n", stderr); \
	fprintf(stderr, " pixels    = %d\n",  n_pixels);      \
	fprintf(stderr, " pixels[0] = %lu\n", pixels[0]);     \
	fprintf(stderr, " reserve   = %d\n",  n_reserve);     \
	fprintf(stderr, " planes    = %d\n",  n_planes);      \
	fprintf(stderr, " colors    = %d\n",  n_colors);
#else
#define PRINT_MAP_DEBUG
#endif


void create_colormap(Display *display, Colormap *colormap, Pixel *pixels_return,
		     int     n_pixels)
{
    unsigned long planes[PLANES_MAX];
    unsigned long pixels[COLORS_MAX];
    XColor        colors[COLORS_MAX];
    XVisualInfo   visinfo;
    int           n_colors, n_reserve, n_planes;
    int           i;

    if (!XMatchVisualInfo(display, DefaultScreen(display), 8, PseudoColor, &visinfo))
    {
	fputs("create_colormap: Need PseudoColor display to create r/w colormap.\n", stderr);
	exit(-1);
    }

    n_colors  = (visinfo.colormap_size < COLORS_MAX) ? visinfo.colormap_size : COLORS_MAX;
    n_planes  = ln_2(n_pixels);
    n_reserve = n_colors - pow_2(n_planes);

    if (n_reserve < 0)
    {
	fputs("create_colormap: Number of pixels to high.\n", stderr);
	PRINT_MAP_DEBUG;
	exit(1);
    }

    /*
     * Query existing colors of the default colormap
     */
    for (i = 0; i < n_reserve; i++)
    {
	colors[i].pixel = i;
	colors[i].flags = DoRed | DoGreen | DoBlue;
    }
    XQueryColors(display, DefaultColormap(display, visinfo.screen), colors, n_reserve);

    /*
     * Create a new colormap, create and store existing entries
     */
    *colormap = XCreateColormap( display,
				 DefaultRootWindow(display),
				 visinfo.visual,
				 AllocNone );
    /*
     * Copy n_reserve of exisiting colors into the new colormap
     */
    if (!XAllocColorCells(display, *colormap, True, NULL, 0, pixels, n_reserve))
    {
	fputs("create_colormap: Can't allocate cells for existing colors.\n", stderr);
	PRINT_MAP_DEBUG;
	exit(1);
    }
    XStoreColors(display, *colormap, colors, n_reserve);

    /*
     * Now allocate 2^(n_planes) color pixels (entries)
     */
    if (!XAllocColorCells(display, *colormap, True, planes, n_planes, pixels, 1))
    {
	fputs("create_colormap: Can't allocate color planes and cells.\n", stderr);
	PRINT_MAP_DEBUG;
	exit(1);
    }

    PRINT_MAP_DEBUG;

    for (i = 0; i < n_pixels; i++)
	pixels_return[i] = pixels[i];
}

/*************************************************************************/

